/* Copyright (C) 2003 David Fries <dfries@mail.win.org>
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation
 * $Id: ppmaverage.cc,v 1.7 2003/12/15 04:13:39 david Exp $
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <dlfcn.h>

#include "OpenEXRWriter.h"

#define READERROR(file) \
{\
	fprintf(stderr, "Error reading %s: %s\n", file, strerror(errno));\
	exit(1);\
}

/* Open the file and read the header information, verify that it is a
 * ppm format image, then fill out the width and height as well as place
 * the file pointer at the beginning of the data.
 */
void ReadHeader(const char * filename, int &fd, int &width, int &height)
{
	fd = open(filename, O_RDONLY);
	if(fd == -1)
	{
		fprintf(stderr, "Error opening file %s: %s\n", filename,
			strerror(errno));
		exit(1);
	}
	char buf[80] = {0};
	int result = read(fd, buf, sizeof(buf));
	if(result == -1)
		READERROR(filename);
	if(result != sizeof(buf))
		READERROR(filename);
	buf[sizeof(buf)-1] = 0;
	if(strncmp("P6\n", buf, 3))
	{
		fprintf(stderr, "%s isn't a ppm file\n", filename);
		exit(1);
	}
	width = 0, height = 0;
	sscanf(buf+3, "%d %d", &width, &height);
	if(!width || !height)
	{
		fprintf(stderr, "Bad image size for %s\n", filename);
		exit(1);
	}
	const char * check = buf+3;
	// find the second new line, we are already past the first
	while(*check)
		if(*check++ == '\n')
			break;
	
	if(strncmp("255\n", check, 4))
	{
		fprintf(stderr, "Only know how to deal with ppm files with "
			"0-255 pixel values\n");
		exit(1);
	}
	check+=4;

	result = lseek(fd, check-buf, SEEK_SET);
	if(result == -1)
	{
		fprintf(stderr, "Error seeking in the file %s: %s\n", filename,
			strerror(errno));
		exit(1);
	}
}

/* Write a PPM format image from the given data which is a sum of 
 * incount images.  image must be large enough to hold the final image,
 * the memory is presumed to be around and not needed at this point in the
 * program so no need to allocate separate memory.
 */
void WritePPMImage(const char * outfile, unsigned short *data,
	unsigned char *image, int incount, int width, int height)
{
	// open and then write out the new image
	int result, fd, total=width*height*3;
	fd = open(outfile, O_WRONLY|O_CREAT|O_EXCL, 0644);
	if(fd == -1)
	{
		fprintf(stderr, "Error opening output image %s: %s\n",
			outfile, strerror(errno));
		exit(1);
	}
	FILE *outstream = fdopen(fd, "w");
	if(!outstream)
	{
		fprintf(stderr, "Error creating file output stream\n");
		exit(1);
	}
	fprintf(outstream, "P6\n%d %d\n255\n", width, height);
	fflush(outstream);
	result = write(fd, image, total);
	if(result == -1)
	{
		fprintf(stderr, "Error writing file %s: %s\n", outfile,
			strerror(errno));
		exit(1);
	}
	if(result != total)
	{
		fprintf(stderr, "Short write to %s\n", outfile);
		exit(1);
	}
	fclose(outstream);
}

WriteEXRImageFuncType LoadEXRFunction(const char * programname)
{
	#define OPENEXRLIBRARY "libOpenEXRWriter.so"
	const char * library_name = OPENEXRLIBRARY;
	void *library = dlopen(library_name, RTLD_NOW);
	char * errorMessage;
	if(!library)
	{
		errorMessage = strdup(dlerror());
		const char * absName = COMPILE_DIRECTORY "/" OPENEXRLIBRARY;
		library = dlopen(absName, RTLD_NOW);
		if(!library)
		{
			fprintf(stderr, "Error loading required library "
				"%s\n", errorMessage);
			fprintf(stderr, "Also failed when using %s\n",
				dlerror());
			exit(2);
		}
		free(errorMessage);
	}
	const char *symbolName = "WriteEXRImage";
	WriteEXRImageFuncType function = (WriteEXRImageFuncType)
		dlsym(library, symbolName);
	errorMessage = dlerror();
	if(errorMessage)
	{
		fprintf(stderr, "Error reading symbol %s: %s\n", symbolName,
			errorMessage);
		exit(1);
	}
	return function;
	#undef OPENEXRLIBRARY
}

void Usage(const char * program)
{
	fprintf(stderr, "Usage: %s infile1.ppm [ ... infilen.ppm ] "
		"-o [ outfile.ppm | outfile.exr ]\n"
		"\tn <= 256, .exr is the OpenEXR format developed by\n"
		"\tIndustrial Light & Magic, supporting 32 bit floating\n"
		"\tpoint values per color\n"
		"\tAll input files must have the same width and height\n"
		"\tThis program averages the color values to correct for \n"
		"\tnoise so any motion is a bad idea.\n", program);
	exit(1);
}

int main(int argc, char ** argv)
{
	if(argc < 4)
		Usage(argv[0]);
	const char *infile[argc], *outfile=NULL;
	enum OutputFormat { IMAGE_PPM, IMAGE_EXR } OutputImage = IMAGE_PPM;
	WriteEXRImageFuncType WriteEXRImageFunction = NULL;
	// read in the file names
	for(int i=1; i<argc; i++)
	{
		if(!strcmp("-o", argv[i]))
		{
			outfile = argv[i+1];
			if(i+1>=argc || i+2<argc)
			{
				fprintf(stderr, "Error, -o can only be listed "
					"once and be the second to last "
					"argument\n");
				Usage(argv[0]);
			}
			// look for the file extension
			const char * ext = outfile;
			while(*ext)
			{
				if(*ext++ == '.')
				{
					if(!strcmp(ext, "ppm"))
						OutputImage = IMAGE_PPM;
					else if(!strcmp(ext, "exr"))
					{
						OutputImage = IMAGE_EXR;
						WriteEXRImageFunction =
							LoadEXRFunction(
								argv[0]);
					}
					else
					{
						fprintf(stderr, "Unknown image "
							"format %s\n", ext);
						exit(1);
					}
				}
			}
			break;
		}
		infile[i-1] = argv[i];
	}
	int width=0, height=0;
	unsigned short incount = argc-3;
	if(incount >= 256)
	{
		fprintf(stderr, "Sorry, only 256 input images are allowed\n");
		Usage(argv[0]);
	}
	// As long as we don't have more than 256 images we don't overfloat
	// an unsigned short.  This way we don't loose any precision.
	unsigned short *data = NULL;
	unsigned char *image = NULL;

	int fd, tmpWidth, tmpHeight, result;
	int total=0;
	for( unsigned short i=0; i<incount; i++)
	{
		ReadHeader(infile[i], fd, tmpWidth, tmpHeight);
		if(!data)
		{
			// allocate the data after opening the first image
			// since we now know the image size
			width = tmpWidth;
			height = tmpHeight;
			total = width*height*3;
			data = new unsigned short[total];
			// start the average at zero
			memset(data, 0, total*sizeof(unsigned short));
			image = new unsigned char[total];
		}
		if(width!=tmpWidth || height!=tmpHeight)
		{
			fprintf(stderr, "All images must be the same width "
				"and height, %s is %dx%d, expected %dx%d\n",
				infile[i], tmpWidth, tmpHeight, width, height);
			exit(1);
		}
		// read in the image
		result = read(fd, image, total);
		if(result == -1)
			READERROR(infile[i]);
		if(result != total)
		{
			fprintf(stderr, "Short read of %s\n", infile[i]);
			exit(1);
		}

		// add this image to the cululated values
		for(int p=0; p<total; p++)
			data[p] += (unsigned short)image[p];

		close(fd);
	}
	// calculate the average image
	for(int p=0; p<total; p++)
		image[p] = (unsigned char)(data[p]/incount);
	
	switch(OutputImage)
	{
	case IMAGE_PPM:
		WritePPMImage(outfile, data, image, incount, width, height);
		break;
	case IMAGE_EXR:
		WriteEXRImageFunction(outfile, data, incount, width, height);
		break;
	default:
		fprintf(stderr, "Error unknown image file type\n");
	}

	delete[]image;
	delete[]data;
	
	return 0;
}
